import os
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"

from .game import Game
from .tagged import Tagged
from .constants import Constants
from .config import Config
from .methods import Methods
